"""The 'worktoy.attr' module implements the descriptor protocol."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ._field import Field
from ._attri_box import AttriBox

__all__ = [
    'Field',
    'AttriBox',
]
