"""
Provides the example from the TypeSig documentation.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ._type_sig_example import main

__all__ = [
    "main",
]
